# Copyright (C) 2025 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import typing

from swh.core.statsd import Statsd, TimedContextManagerDecorator
from swh.fuse import LOGGER_NAME
from swh.model.swhids import CoreSWHID
from swh.objstorage.factory import get_objstorage
from swh.objstorage.interface import objid_from_dict
from swh.storage import get_storage

if typing.TYPE_CHECKING:
    from swh.objstorage.interface import ObjStorageInterface
    from swh.storage.interface import StorageInterface

from . import ContentBackend


class ObjStorageBackend(ContentBackend):
    """
    This content backend relies on an ``swh-storage`` service,
    and maybe an ``swh-objstorage`` directly.

    See :ref:`Configuring files' download <swh-fuse-config-file-download>`.
    """

    def __init__(self, conf: dict):
        self.logger = logging.getLogger(LOGGER_NAME)

        try:
            self.storage: StorageInterface = get_storage(**conf["content"]["storage"])
        except KeyError:
            raise ValueError(
                "`content` configuration block should contain at least a `storage` object"
            )

        try:
            self.objstorage: ObjStorageInterface | None = get_objstorage(
                **conf["content"]["objstorage"]
            )
        except KeyError:
            self.objstorage = None

        self.statsd = Statsd()
        self.storage_tracker = TimedContextManagerDecorator(
            self.statsd, "swhfuse_waiting_storage"
        )
        self.objstorage_tracker = TimedContextManagerDecorator(
            self.statsd, "swhfuse_waiting_objstorage"
        )

    def shutdown(self) -> None:
        self.logger.info(
            "Spent %f ms waiting for storage",
            self.storage_tracker.total_elapsed,
        )
        self.logger.info(
            "Spent %f ms waiting for objstorage backend",
            self.objstorage_tracker.total_elapsed,
        )

    async def get_blob(self, swhid: CoreSWHID) -> bytes:
        """
        Fetch the content of a ``cnt`` object.
        """
        self.statsd.increment("swhfuse_get_blob")
        hashes = None
        try:
            with self.storage_tracker:
                found = self.storage.content_get([swhid.object_id], algo="sha1_git")
                if found and found[0]:
                    hashes = objid_from_dict(found[0].hashes())
            if hashes:
                if self.objstorage is None:
                    self.logger.debug("downloading %s from storage", hashes)
                    with self.storage_tracker:
                        content = self.storage.content_get_data(hashes)
                        if content:
                            return content
                    raise ValueError(f"SWH-storage cannot get object {hashes}")
                # else let's reach the "try objstorage" block
            else:
                raise ValueError(f"SWH-storage cannot find object {swhid}")
        except BaseException as e:
            self.statsd.increment("swhfuse_blob_not_in_storage")
            self.logger.error("Failed to fetch %s from storage: %s", swhid, e)
            raise

        try:
            self.logger.debug("downloading %s from objstorage", hashes)
            with self.objstorage_tracker:
                return self.objstorage.get(hashes)
            raise ValueError(f"SWH-objstorage cannot find object {hashes}")

        except BaseException as e:
            self.statsd.increment("swhfuse_blob_not_in_objstorage")
            self.logger.error("Failed to fetch %s from objstorage: %s", swhid, e)
            raise



# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import asyncio
import base64
from datetime import date, datetime, timedelta, timezone
import hashlib
import logging
from time import sleep
from typing import Any, Dict, List, Tuple
from urllib.parse import unquote_plus

from google.protobuf.field_mask_pb2 import FieldMask
import grpc
from grpc_health.v1 import health_pb2, health_pb2_grpc

from swh.core.statsd import Statsd, TimedContextManagerDecorator
from swh.fuse import LOGGER_NAME
import swh.graph.grpc.swhgraph_pb2 as swhgraph
import swh.graph.grpc.swhgraph_pb2_grpc as swhgraph_grpc
from swh.model.swhids import CoreSWHID, ObjectType

from . import GraphBackend


def decode_or_base64(data: bytes) -> str:
    """
    If ``data`` is not proper UTF-8, return it as base64.
    """
    try:
        return data.decode()
    except UnicodeDecodeError:
        return base64.b64encode(data).decode()


class CompressedGraphBackend(GraphBackend):
    """
    A Backend querying a compressed graph instance via gRPC.
    """

    def __init__(self, conf: dict):
        """
        Only needs ``graph.grpc-url``.
        """
        self.grpc_channel = grpc.insecure_channel(
            conf["graph"]["grpc-url"],
            [
                # advanced options are listed in
                # github.com/grpc/grpc/blob/master/include/grpc/impl/channel_arg_names.h
                #
                # disable the limit on message length, because listing some folders
                # may exceed the default value
                ("grpc.max_receive_message_length", -1),
            ],
        )
        self.grpc_stub = swhgraph_grpc.TraversalServiceStub(self.grpc_channel)
        self.logger = logging.getLogger(LOGGER_NAME)
        self.time_tracker = TimedContextManagerDecorator(
            Statsd(), "swhfuse_waiting_graph"
        )
        self._check_connectivity()

    def _check_connectivity(self):
        """
        Connection errors have been observed, especially when spawning many swh.fuse
        instances against a single swh.graph.grpc-server over HPC. This pre-mount check
        avoids showing a mountpoint that's not yet ready.
        """
        health_stub = health_pb2_grpc.HealthStub(self.grpc_channel)
        for i in range(30):
            try:
                request = health_pb2.HealthCheckRequest(
                    service="swh.graph.TraversalService"
                )
                resp = health_stub.Check(request)
                if resp.status == health_pb2.HealthCheckResponse.SERVING:
                    if i > 0:
                        self.logger.info(
                            "Received a positive health-check from the graph server "
                            "only after %d attempts",
                            i + 1,
                        )
                    return
            except grpc.RpcError as rpc_error:
                if rpc_error.code() != grpc.StatusCode.UNAVAILABLE:
                    raise
            sleep(1)
        raise RuntimeError("swh-graph-grpc-server does not seem active.")

    def shutdown(self) -> None:
        self.grpc_channel.close()
        self.logger.info(
            "Spent %f ms waiting for graph server",
            self.time_tracker.total_elapsed,
        )

    async def get_metadata(self, swhid: CoreSWHID) -> Dict | List:
        loop = asyncio.get_event_loop()
        with self.time_tracker:
            raw = await loop.run_in_executor(
                None,
                self.grpc_stub.GetNode,
                swhgraph.GetNodeRequest(swhid=str(swhid)),
            )

        match swhid.object_type:
            case ObjectType.SNAPSHOT:
                return self._snapshot_metadata(raw)

            case ObjectType.REVISION:
                return self._revision_metadata(swhid, raw)

            case ObjectType.RELEASE:
                return self._release_metadata(swhid, raw)

            case ObjectType.DIRECTORY:
                return await self._directory_metadata(swhid, raw)

            case ObjectType.CONTENT:
                return await self._content_metadata(swhid, raw)

            case _:
                raise NotImplementedError(
                    f"get_metadata({swhid.object_type}) not supported"
                )

    def _snapshot_metadata(self, raw: swhgraph.Node) -> Dict:
        metadata = {}
        for successor in raw.successor:
            target = CoreSWHID.from_string(successor.swhid)
            for branch in successor.label:
                branch_name = decode_or_base64(branch.name)
                metadata[branch_name] = {
                    "target": target.object_id.hex(),
                    "target_type": target.object_type.name.lower(),
                }
        return metadata

    def _revision_metadata(self, swhid: CoreSWHID, raw: swhgraph.Node) -> Dict:
        parents = []
        directory = None
        for successor in raw.successor:
            target = CoreSWHID.from_string(successor.swhid)
            if target.object_type == ObjectType.DIRECTORY:
                directory = target.object_id.hex()
            elif target.object_type == ObjectType.REVISION:
                parent = CoreSWHID.from_string(successor.swhid)
                parents.append({"id": parent.object_id.hex()})
            else:
                raise ValueError(
                    f"Unsupported successor type for {swhid}: {target.object_type}"
                )
        # we also provide fields from protobuf message RevisionData
        try:
            author_date = datetime.fromtimestamp(raw.rev.author_date, tz=timezone.utc)
            author_tz = timezone(timedelta(minutes=raw.rev.author_date_offset))
            author_date = author_date.astimezone(author_tz)
            author_date_str = author_date.isoformat()
        except ValueError:
            author_date_str = None

        try:
            committer_date = datetime.fromtimestamp(
                raw.rev.committer_date, tz=timezone.utc
            )
            committer_tz = timezone(timedelta(minutes=raw.rev.committer_date_offset))
            committer_date = committer_date.astimezone(committer_tz)
            committer_date_str = committer_date.isoformat()
        except ValueError:
            committer_date_str = None

        message = decode_or_base64(raw.rev.message)
        metadata = {
            "author": raw.rev.author,
            "committer": raw.rev.committer,
            "message": message,
            "date": author_date_str,
            "committer_date": committer_date_str,
            "parents": parents,
            "directory": directory,
            "id": swhid.object_id.hex(),
        }
        return metadata

    def _release_metadata(self, swhid: CoreSWHID, raw: swhgraph.Node) -> Dict:
        for successor in raw.successor:
            target = CoreSWHID.from_string(successor.swhid)
            break
        else:
            raise ValueError(f"Cannot find target for release {swhid}")

        try:
            rel_date = datetime.fromtimestamp(raw.rel.author_date, tz=timezone.utc)
            rel_tz = timezone(timedelta(minutes=raw.rel.author_date_offset))
            rel_date = rel_date.astimezone(rel_tz)
            rel_date_str = rel_date.isoformat()
        except ValueError:
            rel_date_str = None

        message = decode_or_base64(raw.rel.message)
        name = decode_or_base64(raw.rel.name)

        metadata = {
            "target": target.object_id.hex(),
            "target_type": target.object_type.name.lower(),
            "id": swhid.object_id.hex(),
            "message": message,
            "name": name,
            "author": raw.rel.author,
            "date": rel_date_str,
        }
        return metadata

    async def _directory_metadata(self, swhid: CoreSWHID, raw: swhgraph.Node) -> List:
        """
        In this case the ``raw`` object obtained from ``getNode`` is not enough,
        because we also need each successors' ``length`` property (if it's a content) to
        be on par with the WebAPI. We use a filtered traversal request to obtain lengths
        in a single gRPC call.
        """

        loop = asyncio.get_event_loop()
        with self.time_tracker:
            raw_cnt = await loop.run_in_executor(
                None,
                self.grpc_stub.Traverse,
                swhgraph.TraversalRequest(
                    src=[str(swhid)],
                    max_depth=1,
                    edges="dir:cnt",
                    mask=FieldMask(paths=["swhid", "cnt"]),
                ),
            )
        cnt_metadata = {}
        for item in raw_cnt:
            if item.HasField("cnt"):
                cnt_metadata[item.swhid] = {
                    "length": item.cnt.length,
                    "status": "skipped" if item.cnt.is_skipped else "visible",
                }

        metadata = []
        for successor in raw.successor:
            target = CoreSWHID.from_string(successor.swhid)
            target_type = target.object_type.value
            if target_type == ObjectType.CONTENT.value:
                # complying with swh-web-client.typify_json is hard
                target_type = "file"
            for label in successor.label:
                name = decode_or_base64(label.name)
                entry = {
                    "dir_id": swhid.object_id.hex(),
                    "name": name,
                    "perms": label.permission,
                    "type": target_type,
                    "target": target.object_id.hex(),
                }
                if target.object_type == ObjectType.CONTENT:
                    if successor.swhid in cnt_metadata:
                        entry.update(cnt_metadata[successor.swhid])
                    else:
                        self.logger.warning(
                            "%s listed as successor of %s, but not fetched by raw_cnt",
                            target,
                            swhid,
                        )
                metadata.append(entry)

        return metadata

    async def _content_metadata(self, swhid: CoreSWHID, raw: swhgraph.Node) -> Dict:
        return {
            "length": raw.cnt.length,
            "status": "skipped" if raw.cnt.is_skipped else "visible",
        }

    async def get_history(self, swhid: CoreSWHID) -> List[Tuple[str, str]]:
        loop = asyncio.get_event_loop()
        with self.time_tracker:
            raw_cnt = await loop.run_in_executor(
                None,
                self.grpc_stub.Traverse,
                swhgraph.TraversalRequest(
                    src=[str(swhid)],
                    edges="rev:rev",
                    mask=FieldMask(paths=["swhid"]),
                ),
            )
        child_str = str(swhid)
        result = []
        for entry in raw_cnt:
            entry_str = str(entry.swhid)
            result.append((child_str, entry_str))
            child_str = entry_str
        return result

    async def get_visits(self, url_encoded: str) -> List[Dict[str, Any]]:
        url = unquote_plus(url_encoded)
        swhid = "swh:1:ori:" + hashlib.sha1(url.encode()).hexdigest()

        loop = asyncio.get_event_loop()
        with self.time_tracker:
            ori_node = await loop.run_in_executor(
                None,
                self.grpc_stub.GetNode,
                swhgraph.GetNodeRequest(swhid=str(swhid)),
            )

        origin = url
        visits = []
        for entry in ori_node.successor:
            snapshot = CoreSWHID.from_string(entry.swhid)
            for label in entry.label:
                visit_date = date.fromtimestamp(label.visit_timestamp)
                visits.append(
                    {
                        "date": visit_date.isoformat(),
                        "origin": origin,
                        "snapshot": snapshot.object_id.hex(),
                    }
                )

        return visits



# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.core.config import merge_configs
from swh.fuse.cli import DEFAULT_CONFIG
from swh.fuse.fuse import SwhFsTmpMount


def test_swh_fuse_context():
    config = merge_configs(
        DEFAULT_CONFIG,
        {
            "cache": {
                "metadata": {"in-memory": True},
                "blob": {"bypass": True},
                "direntry": {"maxram": "10%"},
            },
        },
    )
    with SwhFsTmpMount(config=config) as mountpoint:
        cache = mountpoint / "cache"
        assert cache.is_dir()
        subfolders = list(cache.glob("*"))
        assert subfolders == [cache / "origin"]
    assert not mountpoint.is_dir()



# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
from pathlib import Path


def test_revision(fuse_graph_mountpoint: Path, example_revision: str):
    root = fuse_graph_mountpoint / "archive" / example_revision
    assert root.is_dir()
    assert (root / "meta.json").is_symlink()

    with (root / "meta.json").open() as f:
        parsed = json.loads(f.read())
        assert parsed["id"] in example_revision
        assert "author" in parsed
        assert "committer" in parsed
        assert "message" in parsed
        assert "parents" in parsed
        assert "directory" in parsed
        assert parsed["date"] == "2005-03-18T13:14:00+02:00"
        assert parsed["committer_date"] == "2005-03-18T16:19:10+02:00"

    assert (root / "parents").is_dir()
    assert (root / "parent").is_symlink()
    assert (root / "root").is_symlink()

    assert (root / "history").is_dir()
    # TODO



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import asyncio
import json
import logging
import os
from pathlib import Path
from tempfile import TemporaryDirectory
import threading
import time

from click.testing import CliRunner
import pytest
import requests_mock

from swh.fuse import LOGGER_NAME, fuse
import swh.fuse.cli as cli

from .data.api_data import API_URL, MOCK_ARCHIVE


@pytest.fixture(scope="module")
def fuse_mntdir():
    with (
        TemporaryDirectory(suffix=".swh-fuse-test") as tmpdir,
        requests_mock.Mocker() as mocker,
    ):
        for api_call, data in MOCK_ARCHIVE.items():
            # Convert Python dict JSON into a string (only for non-raw API call)
            if not api_call.endswith("raw/") and not api_call.startswith("graph/"):
                data = json.dumps(data)
            mocked_method = mocker.get
            if api_call.startswith("origin/") and api_call.endswith("get/"):
                mocked_method = mocker.head
            mocked_method(f"{API_URL}/{api_call}", text=data)

        mountpoint = Path(tmpdir)
        config = {
            "cache": {
                "metadata": {"in-memory": True},
                "blob": {"in-memory": True},
                "direntry": {"maxram": "10%"},
            },
            "web-api": {"url": API_URL, "auth-token": None},
            "json-indent": None,
        }

        def mount():
            loop = asyncio.new_event_loop()
            logging.getLogger(LOGGER_NAME).setLevel(logging.DEBUG)
            try:
                loop.run_until_complete(fuse.main([], mountpoint, config))
            finally:
                if loop.is_running():
                    loop.stop()
                loop.close()

        mount_thread = threading.Thread(target=mount)
        mount_thread.start()

        for i in range(30):
            try:
                root = os.listdir(tmpdir)
                if root:
                    break
            except FileNotFoundError:
                pass
            time.sleep(0.1)
        else:
            raise FileNotFoundError(f"Could not mount FUSE in {tmpdir}")

        try:
            yield mountpoint
        finally:
            CliRunner().invoke(cli.umount, [tmpdir])



#!/usr/bin/env python3

# Copyright (C) 2020-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


import json
from typing import Any, Dict

import requests

from swh.fuse.tests.web_api.api_url import swhid_to_history_url, swhid_to_web_url
from swh.fuse.tests.web_api.data.config import (
    ALL_ENTRIES,
    FAKE_SNP_SPECIAL_CASES,
    FAKE_SNP_SPECIAL_CASES_SWHID,
    ORIGIN_URL,
    REV_SMALL_HISTORY,
)
from swh.model.hashutil import hash_to_bytes
from swh.model.swhids import CoreSWHID, ObjectType

API_URL_real = "https://archive.softwareheritage.org/api/1"
API_URL_test = "https://invalid-test-only.archive.softwareheritage.org/api/1"

# Use your own API token to lift rate limiting. Note: this is not necessary to generate
# the API data only once but can be useful when re-generating it multiple times.
API_TOKEN = ""


MOCK_ARCHIVE: Dict[str, Any] = {}
# Temporary map (swhid -> metadata) to ease data generation
METADATA: Dict[CoreSWHID, Any] = {}


def get_from_api(endpoint: str) -> str:
    headers = {"Authorization": f"Bearer {API_TOKEN}"} if API_TOKEN else {}
    result = requests.get(f"{API_URL_real}/{endpoint}", headers=headers)
    if result.status_code != 200:
        raise RuntimeError(f"API request failed: {result.status_code} {result.text}")
    return result.text


def generate_archive_web_api(
    swhid: CoreSWHID, raw: bool = False, recursive: bool = False
) -> None:
    # Already in mock archive
    if swhid in METADATA and not raw:
        return

    url = swhid_to_web_url(swhid, raw)

    data = get_from_api(url)
    if not raw:
        data = json.loads(data)

    MOCK_ARCHIVE[url] = data
    METADATA[swhid] = data

    # Retrieve additional needed data for different artifacts (eg: content's
    # blob data, release target, etc.)
    if recursive:
        if swhid.object_type == ObjectType.CONTENT:
            generate_archive_web_api(swhid, raw=True)
        elif swhid.object_type == ObjectType.RELEASE:
            try:
                target_type = METADATA[swhid]["target_type"]
                target_id = METADATA[swhid]["target"]
            except KeyError:
                print(f"keyerror on METADATA[{swhid}]={METADATA[swhid]}")
            target = CoreSWHID(
                object_type=ObjectType[target_type.upper()],
                object_id=hash_to_bytes(target_id),
            )
            generate_archive_web_api(target, recursive=True)


def generate_archive_graph_api(swhid: CoreSWHID) -> None:
    if swhid.object_type == ObjectType.REVISION:
        # Empty history for all revisions (except REV_SMALL_HISTORY used in tests)
        url = swhid_to_history_url(swhid)
        MOCK_ARCHIVE[url] = "[]"
        if str(swhid) == REV_SMALL_HISTORY:
            history = json.loads(get_from_api(url))
            MOCK_ARCHIVE[url] = history
            revs = set(CoreSWHID.from_string("swh:1:rev:" + e["id"]) for e in history)
            for swhid in revs:
                generate_archive_web_api(swhid, recursive=False)


def generate_origin_archive_web_api(url: str):
    url_visits = f"origin/{url}/visits/"
    data = get_from_api(url_visits)
    data = json.loads(data)
    MOCK_ARCHIVE[url_visits] = data
    # Necessary since swh-fuse will check the origin URL using the get/ endpoint
    url_get = f"origin/{url}/get/"
    MOCK_ARCHIVE[url_get] = ""


if __name__ == "__main__":
    for entry in ALL_ENTRIES:
        swhid = CoreSWHID.from_string(entry)
        generate_archive_web_api(swhid, recursive=True)
        generate_archive_graph_api(swhid)

    # Custom fake snapshot to handle most special cases
    MOCK_ARCHIVE[swhid_to_web_url(FAKE_SNP_SPECIAL_CASES_SWHID)] = {
        "branches": FAKE_SNP_SPECIAL_CASES
    }

    # Origin artifacts are not identified by SWHID but using an URL
    generate_origin_archive_web_api(ORIGIN_URL)

    print("# GENERATED FILE, DO NOT EDIT.")
    print("# Run './gen-api-data.py > api_data.py' instead.")
    print("# flake8: noqa")
    print("from typing import Any, Dict")
    print("")
    print(f"API_URL = '{API_URL_test}'\n")
    print(f"MOCK_ARCHIVE: Dict[str, Any] = {MOCK_ARCHIVE}")



import json
import os

import requests_mock

from .test_revision import wait_check_no_status
from .unusual_data import (
    API_URL,
    BROKEN_RELEASE_SWHID,
    BROKEN_REVISION_SWHID,
    MOCK_ARCHIVE,
    REVISION_SWHID,
)


def test_unusual_cases(fuse_mntdir):
    with requests_mock.Mocker() as mocker:
        for api_call, data in MOCK_ARCHIVE.items():
            mocker.get(f"{API_URL}/{api_call}", text=data)

        release_path = fuse_mntdir / "archive" / BROKEN_RELEASE_SWHID

        meta_path = release_path / "meta.json"
        meta_content = json.loads(meta_path.read_text())
        assert meta_content["target_type"] == "revision"
        assert meta_content["target"] in REVISION_SWHID
        assert meta_content["date"] is None

        target_path = release_path / "target"
        expected = set(["meta.json", "root", "parent", "parents", "history"])
        actual = set(os.listdir(target_path))
        assert expected.issubset(actual)

        target_history_bydate = target_path / "history" / "by-date"
        wait_check_no_status(target_history_bydate)

        history = list(os.listdir(target_history_bydate))
        assert history == ["1970"]
        broken_date_path = target_history_bydate / "1970" / "01" / "01"
        broken_rev_path = list(os.listdir(broken_date_path))[0]
        broken_rev_meta = json.loads(
            (broken_date_path / broken_rev_path / "meta.json").read_text()
        )
        assert broken_rev_meta["id"] in BROKEN_REVISION_SWHID



# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from pathlib import Path

from swh.model.model import Content


def test_content(
    fuse_graph_mountpoint: Path, example_content_swhid: str, example_content: Content
):
    root = fuse_graph_mountpoint / "archive" / example_content_swhid

    assert root.is_file()
    with root.open("rb") as f:
        content = f.read()
        assert content == example_content.data
